from typing import Dict, Any, Optional
import uuid

import asyncio
import aiohttp
from python_graphql_client import GraphqlClient

from chainlit.client.base import UserDict

from .base import BaseDBClient, BaseAuthClient, PaginatedResponse, PageInfo, UserDict

from chainlit.logger import logger
from chainlit.config import config


class GraphQLClient:
    def __init__(self, project_id: str, access_token: str):
        self.project_id = project_id
        self.headers = {
            "Authorization": access_token,
            "content-type": "application/json",
        }
        graphql_endpoint = f"{config.chainlit_server}/api/graphql"
        self.graphql_client = GraphqlClient(
            endpoint=graphql_endpoint, headers=self.headers
        )

    def query(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        :param query: The GraphQL query string.
        :param variables: A dictionary of variables for the query.
        :return: The response data as a dictionary.
        """
        return self.graphql_client.execute_async(query=query, variables=variables)

    def check_for_errors(self, response: Dict[str, Any], raise_error: bool = False):
        if "errors" in response:
            if raise_error:
                raise Exception(response["errors"][0])
            logger.error(response["errors"][0])
            return True
        return False

    def mutation(self, mutation: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL mutation.

        :param mutation: The GraphQL mutation string.
        :param variables: A dictionary of variables for the mutation.
        :return: The response data as a dictionary.
        """
        return self.graphql_client.execute_async(query=mutation, variables=variables)


class CloudAuthClient(BaseAuthClient, GraphQLClient):
    def __init__(self, project_id: str, access_token: str):
        super().__init__(project_id, access_token)

    async def get_user_infos(
        self,
    ) -> UserDict:
        data = {"projectId": self.project_id}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}/api/me",
                json=data,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    raise ValueError(f"Failed to get user infos. {r.status}: {reason}")

                json = await r.json()
                self.user_infos = json
                return self.user_infos

    async def is_project_member(self) -> bool:
        try:
            user = await self.get_user_infos()
            return user["role"] != "ANONYMOUS"
        except ValueError as e:
            logger.error(e)
            return False


class CloudDBClient(BaseDBClient, GraphQLClient):
    conversation_id: Optional[str] = None
    lock: asyncio.Lock

    def __init__(self, project_id: str, access_token: str):
        self.lock = asyncio.Lock()
        super().__init__(project_id, access_token)

    async def create_user(self, variables: UserDict) -> bool:
        raise NotImplementedError

    async def get_project_members(self):
        query = """query ($projectId: String!) {
                    projectMembers(projectId: $projectId) {
                    edges {
                        cursor
                        node {
                        role
                        user {
                            email
                            name
                        }
                        }
                    }
                    }
                }"""
        variables = {"projectId": self.project_id}
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        members = []

        for edge in res["data"]["projectMembers"]["edges"]:
            node = edge["node"]
            role = node["role"]
            name = node["user"]["name"]
            email = node["user"]["email"]
            members.append({"role": role, "name": name, "email": email})

        return members

    async def create_conversation(self) -> int:
        # If we run multiple send concurrently, we need to make sure we don't create multiple conversations.
        async with self.lock:
            if self.conversation_id:
                return self.conversation_id

            mutation = """
            mutation ($projectId: String!, $sessionId: String) {
                createConversation(projectId: $projectId, sessionId: $sessionId) {
                    id
                }
            }
            """
            variables = {"projectId": self.project_id}
            res = await self.mutation(mutation, variables)

            if self.check_for_errors(res):
                logger.warning("Could not create conversation.")
                return None

            return int(res["data"]["createConversation"]["id"])

    async def get_conversation_id(self):
        self.conversation_id = await self.create_conversation()

        return self.conversation_id

    async def delete_conversation(self, conversation_id: int):
        mutation = """mutation ($id: ID!) {
    deleteConversation(id: $id) {
      id
    }
  }"""
        variables = {"id": conversation_id}
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_conversation(self, conversation_id: int):
        query = """query ($id: ID!) {
    conversation(id: $id) {
      id
      createdAt
      messages {
        id
        isError
        indent
        author
        content
        waitForAnswer
        humanFeedback
        language
        prompt
        llmSettings
        authorIsUser
        createdAt
      }
      elements {
        id
        conversationId
        type
        name
        url
        display
        language
        size
        forIds
      }
    }
  }"""
        variables = {
            "id": conversation_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["conversation"]

    async def get_conversations(self, pagination, filter):
        query = """query (
        $first: Int
        $projectId: String!
        $cursor: String
        $withFeedback: Int
        $authorEmail: String
        $search: String
    ) {
        conversations(
        first: $first
        cursor: $cursor
        projectId: $projectId
        withFeedback: $withFeedback
        authorEmail: $authorEmail
        search: $search
        ) {
        pageInfo {
            endCursor
            hasNextPage
        }
        edges {
            cursor
            node {
            id
            createdAt
            elementCount
            messageCount
            author {
                name
                email
            }
            messages {
                content
            }
            }
        }
        }
    }"""

        variables = {
            "projectId": self.project_id,
            "first": pagination.first,
            "cursor": pagination.cursor,
            "withFeedback": filter.feedback,
            "authorEmail": filter.authorEmail,
            "search": filter.search,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        conversations = []

        for edge in res["data"]["conversations"]["edges"]:
            node = edge["node"]
            conversations.append(node)

        page_info = res["data"]["conversations"]["pageInfo"]

        return PaginatedResponse(
            pageInfo=PageInfo(
                hasNextPage=page_info["hasNextPage"],
                endCursor=page_info["endCursor"],
            ),
            data=conversations,
        )

    async def set_human_feedback(self, message_id, feedback):
        mutation = """mutation ($messageId: ID!, $humanFeedback: Int!) {
                        setHumanFeedback(messageId: $messageId, humanFeedback: $humanFeedback) {
                            id
                            humanFeedback
                    }
                }"""
        variables = {"messageId": message_id, "humanFeedback": feedback}
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_message(self):
        raise NotImplementedError

    async def create_message(self, variables: Dict[str, Any]) -> int:
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the message.")
            return None

        variables["conversationId"] = c_id

        mutation = """
        mutation ($conversationId: ID!, $author: String!, $content: String!, $language: String, $prompt: String, $llmSettings: Json, $isError: Boolean, $indent: Int, $authorIsUser: Boolean, $waitForAnswer: Boolean, $createdAt: StringOrFloat) {
            createMessage(conversationId: $conversationId, author: $author, content: $content, language: $language, prompt: $prompt, llmSettings: $llmSettings, isError: $isError, indent: $indent, authorIsUser: $authorIsUser, waitForAnswer: $waitForAnswer, createdAt: $createdAt) {
                id
            }
        }
        """
        res = await self.mutation(mutation, variables)
        if self.check_for_errors(res):
            logger.warning("Could not create message.")
            return None

        return int(res["data"]["createMessage"]["id"])

    async def update_message(self, message_id: int, variables: Dict[str, Any]) -> bool:
        mutation = """
        mutation ($messageId: ID!, $author: String!, $content: String!, $language: String, $prompt: String, $llmSettings: Json) {
            updateMessage(messageId: $messageId, author: $author, content: $content, language: $language, prompt: $prompt, llmSettings: $llmSettings) {
                id
            }
        }
        """
        variables["messageId"] = message_id
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not update message.")
            return False

        return True

    async def delete_message(self, message_id: int) -> bool:
        mutation = """
        mutation ($messageId: ID!) {
            deleteMessage(messageId: $messageId) {
                id
            }
        }
        """
        res = await self.mutation(mutation, {"messageId": message_id})

        if self.check_for_errors(res):
            logger.warning("Could not delete message.")
            return False

        return True

    async def get_element(self, conversation_id, element_id):
        query = """query (
        $conversationId: ID!
        $id: ID!
    ) {
        element(
        conversationId: $conversationId,
        id: $id
        ) {
        id
        conversationId
        type
        name
        url
        display
        language
        size
        forIds
        }
    }"""

        variables = {
            "conversationId": conversation_id,
            "id": element_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["element"]

    async def upsert_element(self, variables):
        c_id = await self.get_conversation_id()

        if not c_id:
            logger.warning("Missing conversation ID, could not persist the element.")
            return None

        if "id" in variables:
            mutation_name = "updateElement"
            mutation = """
            mutation ($conversationId: ID!, $id: ID!, $forIds: [String!]!) {
                updateElement(conversationId: $conversationId, id: $id, forIds: $forIds) {
                    id,
                }
            }
            """
            variables["conversationId"] = c_id
            res = await self.mutation(mutation, variables)
        else:
            mutation_name = "createElement"
            mutation = """
            mutation ($conversationId: ID!, $type: String!, $url: String!, $name: String!, $display: String!, $forIds: [String!]!, $size: String, $language: String) {
                createElement(conversationId: $conversationId, type: $type, url: $url, name: $name, display: $display, size: $size, language: $language, forIds: $forIds) {
                    id,
                    type,
                    url,
                    name,
                    display,
                    size,
                    language,
                    forIds
                }
            }
            """
            variables["conversationId"] = c_id
            res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not persist element.")
            return None

        return res["data"][mutation_name]

    async def upload_element(self, content: bytes, mime: str) -> str:
        id = f"{uuid.uuid4()}"
        body = {"projectId": self.project_id, "fileName": id, "contentType": mime}

        path = f"/api/upload/file"

        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{config.chainlit_server}{path}",
                json=body,
                headers=self.headers,
            ) as r:
                if not r.ok:
                    reason = await r.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return ""
                json_res = await r.json()

        upload_details = json_res["post"]
        permanent_url = json_res["permanentUrl"]

        form_data = aiohttp.FormData()

        # Add fields to the form_data
        for field_name, field_value in upload_details["fields"].items():
            form_data.add_field(field_name, field_value)

        # Add file to the form_data
        form_data.add_field("file", content, content_type="multipart/form-data")
        async with aiohttp.ClientSession() as session:
            async with session.post(
                upload_details["url"],
                data=form_data,
            ) as upload_response:
                if not upload_response.ok:
                    reason = await upload_response.text()
                    logger.error(f"Failed to upload file: {reason}")
                    return ""

                url = f'{upload_details["url"]}/{upload_details["fields"]["key"]}'
                return permanent_url



from typing import List, Dict, Union
from abc import ABC, abstractmethod
import uuid
import asyncio
from datetime import datetime, timezone

from chainlit.telemetry import trace_event
from chainlit.context import get_emitter
from chainlit.config import config
from chainlit.types import (
    LLMSettings,
    AskSpec,
    AskFileSpec,
    AskFileResponse,
    AskResponse,
)
from chainlit.element import Element
from chainlit.action import Action
from chainlit.logger import logger


class MessageBase(ABC):
    id: int = None
    temp_id: str = None
    streaming = False
    created_at: int = None
    fail_on_persist_error: bool = True

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.temp_id = uuid.uuid4().hex
        self.created_at = datetime.now(timezone.utc).isoformat()
        self.emitter = get_emitter()

    @abstractmethod
    def to_dict(self):
        pass

    async def _create(self):
        msg_dict = self.to_dict()
        if self.emitter.db_client and not self.id:
            try:
                self.id = await self.emitter.db_client.create_message(msg_dict)
                if self.id:
                    msg_dict["id"] = self.id
            except Exception as e:
                if self.fail_on_persist_error:
                    raise e
                logger.error(f"Failed to persist message: {str(e)}")

        return msg_dict

    async def update(
        self,
        author: str = None,
        content: str = None,
        language: str = None,
        prompt: str = None,
        llm_settings: LLMSettings = None,
    ):
        """
        Update a message already sent to the UI.
        """
        trace_event("update_message")

        if author:
            self.author = author
        if content:
            self.content = content
        if language:
            self.language = language
        if prompt:
            self.prompt = prompt
        if llm_settings:
            self.llmSettings = llm_settings

        msg_dict = self.to_dict()

        if self.emitter.db_client and self.id:
            await self.emitter.db_client.update_message(self.id, msg_dict)

        await self.emitter.update_message(msg_dict)

        return True

    async def remove(self):
        """
        Remove a message already sent to the UI.
        This will not automatically remove potential nested messages and could lead to undesirable side effects in the UI.
        """
        trace_event("remove_message")

        if self.emitter.db_client and self.id:
            await self.emitter.db_client.delete_message(self.id)

        await self.emitter.delete_message(self.to_dict())

        return True

    async def send(self) -> Union[str, int]:
        if self.content is None:
            self.content = ""

        msg_dict = await self._create()

        if self.streaming:
            self.streaming = False

        await self.emitter.send_message(msg_dict)

        return self.id or self.temp_id

    async def stream_token(self, token: str, is_sequence=False):
        """
        Sends a token to the UI. This is useful for streaming messages.
        Once all tokens have been streamed, call .send() to end the stream and persist the message if persistence is enabled.
        """

        if not self.streaming:
            self.streaming = True
            msg_dict = self.to_dict()
            await self.emitter.stream_start(msg_dict)

        if is_sequence:
            self.content = token
        else:
            self.content += token

        await self.emitter.send_token(
            id=self.id or self.temp_id, token=token, is_sequence=is_sequence
        )


class Message(MessageBase):
    """
    Send a message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (str): The content of the message.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        prompt (str, optional): The prompt used to generate the message. If provided, enables the prompt playground for this message.
        llm_settings (LLMSettings, optional): Settings of the LLM used to generate the prompt. This is useful for debug purposes in the prompt playground.
        language (str, optional): Language of the code is the content is code. See https://react-code-blocks-rajinwonderland.vercel.app/?path=/story/codeblock--supported-languages for a list of supported languages.
        indent (int, optional): If positive, the message will be nested in the UI.
        actions (List[Action], optional): A list of actions to send with the message.
        elements (List[Element], optional): A list of elements to send with the message.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        prompt: str = None,
        llm_settings: LLMSettings = None,
        language: str = None,
        indent: int = 0,
        actions: List[Action] = [],
        elements: List[Element] = [],
    ):
        self.content = content
        self.author = author
        self.prompt = prompt
        self.language = language
        self.indent = indent
        self.actions = actions
        self.elements = elements
        self.llmSettings = None

        if llm_settings is None and prompt is not None:
            self.llmSettings = LLMSettings().to_dict()

        if llm_settings:
            self.llmSettings = llm_settings.to_dict()

        super().__post_init__()

    def to_dict(self):
        _dict = {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "prompt": self.prompt,
            "llmSettings": self.llmSettings,
            "language": self.language,
            "indent": self.indent,
        }

        if self.id:
            _dict["id"] = self.id

        return _dict

    async def send(self):
        """
        Send the message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_message")
        id = await super().send()

        action_coros = [action.send(for_id=str(id)) for action in self.actions]
        element_coros = [element.send(for_id=str(id)) for element in self.elements]
        all_coros = action_coros + element_coros
        await asyncio.gather(*all_coros)

        return id


class ErrorMessage(MessageBase):
    """
    Send an error message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (str): Text displayed above the upload button.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        indent (int, optional): If positive, the message will be nested in the UI.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        indent: int = 0,
        fail_on_persist_error: bool = False,
    ):
        self.content = content
        self.author = author
        self.indent = indent
        self.fail_on_persist_error = fail_on_persist_error

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "indent": self.indent,
            "isError": True,
        }

    async def send(self):
        """
        Send the error message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_error_message")
        return await super().send()


class AskMessageBase(MessageBase):
    async def remove(self):
        removed = await super().remove()
        if removed:
            await self.emitter.clear_ask()


class AskUserMessage(AskMessageBase):
    """
    Ask for the user input before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The content of the prompt.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        timeout: int = 60,
        raise_on_timeout: bool = False,
    ):
        self.content = content
        self.author = author
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
        }

    async def send(self) -> Union[AskResponse, None]:
        """
        Sends the question to ask to the UI and waits for the reply.
        """
        trace_event("send_ask_user")

        if self.streaming:
            self.streaming = False

        msg_dict = await self._create()

        spec = AskSpec(type="text", timeout=self.timeout)

        res = await self.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        return res


class AskFileMessage(AskMessageBase):
    """
    Ask the user to upload a file before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the file will be uploaded to the cloud storage.

    Args:
        content (str): Text displayed above the upload button.
        accept (Union[List[str], Dict[str, List[str]]]): List of mime type to accept like ["text/csv", "application/pdf"] or a dict like {"text/plain": [".txt", ".py"]}.
        max_size_mb (int, optional): Maximum size per file in MB. Maximum value is 100.
        max_files (int, optional): Maximum number of files to upload. Maximum value is 10.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        accept: Union[List[str], Dict[str, List[str]]],
        max_size_mb=2,
        max_files=1,
        author=config.ui.name,
        timeout=90,
        raise_on_timeout=False,
    ):
        self.content = content
        self.max_size_mb = max_size_mb
        self.max_files = max_files
        self.accept = accept
        self.author = author
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "tempId": self.temp_id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
        }

    async def send(self) -> Union[List[AskFileResponse], None]:
        """
        Sends the message to request a file from the user to the UI and waits for the reply.
        """
        trace_event("send_ask_file")

        if self.streaming:
            self.streaming = False

        msg_dict = await self._create()

        spec = AskFileSpec(
            type="file",
            accept=self.accept,
            max_size_mb=self.max_size_mb,
            max_files=self.max_files,
            timeout=self.timeout,
        )

        res = await self.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        if res:
            return [AskFileResponse(**r) for r in res]
        else:
            return None



from fastapi import Request

from chainlit.config import config
from chainlit.client.base import BaseDBClient, BaseAuthClient, UserDict
from chainlit.client.local import LocalAuthClient, LocalDBClient
from chainlit.client.cloud import CloudAuthClient, CloudDBClient
from chainlit.telemetry import trace_event


async def get_auth_client(authorization: str) -> BaseAuthClient:
    # Check authorization
    if not config.project.public and not authorization:
        # Refuse connection if the app is private and no access token is provided
        trace_event("no_access_token")
        raise ConnectionRefusedError("No access token provided")
    elif authorization and config.project.id:
        # Create the auth cloud client
        auth_client = CloudAuthClient(
            project_id=config.project.id,
            access_token=authorization,
        )
        # Check if the user is a member of the project
        is_project_member = await auth_client.is_project_member()
        if not is_project_member:
            raise ConnectionRefusedError("User is not a member of the project")

        return auth_client

    # Default to local auth client
    return LocalAuthClient()


async def get_db_client(
    authorization: str = None, user_infos: UserDict = None
) -> BaseDBClient:
    # Create the database client
    if config.project.database == "cloud":
        if not config.project.id:
            raise ValueError("Project id is required for database mode 'cloud'")

        return CloudDBClient(
            project_id=config.project.id,
            access_token=authorization,
        )
    elif config.project.database == "local":
        return LocalDBClient(user_infos=user_infos)
    elif config.project.database == "custom":
        if not config.code.client_factory:
            raise ValueError("Client factory not provided")

        custom_db_client = await config.code.client_factory(user_infos)
        return custom_db_client


async def get_auth_client_from_request(
    request: Request,
) -> BaseAuthClient:
    authorization = request.headers.get("Authorization")

    # Get the auth client
    auth_client = await get_auth_client(authorization)

    return auth_client


async def get_db_client_from_request(
    request: Request,
) -> BaseDBClient:
    authorization = request.headers.get("Authorization")

    # Get the auth client
    auth_client = await get_auth_client(authorization)

    # Get the db client
    db_client = await get_db_client(authorization, auth_client.user_infos)

    return db_client



import contextvars
from typing import TYPE_CHECKING
from asyncio import AbstractEventLoop

if TYPE_CHECKING:
    from chainlit.emitter import ChainlitEmitter


class ChainlitContextException(Exception):
    def __init__(self, msg="Chainlit context not found", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)


emitter_var = contextvars.ContextVar("emitter")
loop_var = contextvars.ContextVar("loop")


def get_emitter() -> "ChainlitEmitter":
    try:
        return emitter_var.get()
    except LookupError:
        raise ChainlitContextException()


def get_loop() -> AbstractEventLoop:
    try:
        return loop_var.get()
    except LookupError:
        raise ChainlitContextException()



from pydantic.dataclasses import dataclass, Field
from dataclasses_json import dataclass_json
from typing import Dict, List, Union, Any
import uuid
import aiofiles
from io import BytesIO
from enum import Enum
import json
import filetype

from chainlit.context import get_emitter
from chainlit.client.base import BaseDBClient
from chainlit.telemetry import trace_event
from chainlit.types import ElementType, ElementDisplay, ElementSize

mime_types = {
    "text": "text/plain",
    "tasklist": "application/json",
}


@dataclass
class Element:
    # Name of the element, this will be used to reference the element in the UI.
    name: str
    # The type of the element. This will be used to determine how to display the element in the UI.
    type: ElementType
    # Controls how the image element should be displayed in the UI. Choices are “side” (default), “inline”, or “page”.
    display: ElementDisplay = "side"
    # The URL of the element if already hosted somehwere else.
    url: str = None
    # The local path of the element.
    path: str = None
    # The byte content of the element.
    content: bytes = None
    # The ID of the element. This is set automatically when the element is sent to the UI if cloud is enabled.
    id: int = None
    # The ID of the element if cloud is disabled.
    temp_id: str = None
    # The ID of the message this element is associated with.
    for_ids: List[str] = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.emitter = get_emitter()
        self.for_ids = []
        self.temp_id = str(uuid.uuid4())

        if not self.url and not self.path and not self.content:
            raise ValueError("Must provide url, path or content to instantiate element")

    def to_dict(self) -> Dict:
        _dict = {
            "tempId": self.temp_id,
            "type": self.type,
            "url": self.url,
            "name": self.name,
            "display": self.display,
            "size": getattr(self, "size", None),
            "language": getattr(self, "language", None),
            "forIds": getattr(self, "for_ids", None),
        }

        if self.id:
            _dict["id"] = self.id

        return _dict

    async def preprocess_content(self):
        pass

    async def load(self):
        if self.path:
            async with aiofiles.open(self.path, "rb") as f:
                self.content = await f.read()
        else:
            raise ValueError("Must provide path or content to load element")

    async def persist(self, client: BaseDBClient):
        if not self.url and self.content and not self.id:
            # Only guess the mime type when the content is binary
            mime = (
                mime_types[self.type]
                if self.type in mime_types
                else filetype.guess_mime(self.content)
            )
            self.url = await client.upload_element(content=self.content, mime=mime)
        element = await client.upsert_element(self.to_dict())
        return element

    async def before_emit(self, element: Dict) -> Dict:
        return element

    async def remove(self):
        trace_event(f"remove {self.__class__.__name__}")
        await self.emitter.emit("remove_element", {"id": self.id or self.temp_id})

    async def send(self, for_id: str = None):
        element = None

        if not self.content and not self.url and self.path:
            await self.load()

        await self.preprocess_content()

        if for_id:
            self.for_ids.append(for_id)

        # We have a client, persist the element
        if self.emitter.db_client:
            element = await self.persist(self.emitter.db_client)
            self.id = element and element.get("id")

        elif not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        element = self.to_dict()

        element["content"] = self.content

        if self.emitter.emit and element:
            if len(self.for_ids) > 1:
                trace_event(f"update {self.__class__.__name__}")
                await self.emitter.emit(
                    "update_element",
                    {"id": self.id or self.temp_id, "forIds": self.for_ids},
                )
            else:
                trace_event(f"send {self.__class__.__name__}")
                element = await self.before_emit(element)
                await self.emitter.emit("element", element)


@dataclass
class Image(Element):
    type: ElementType = "image"
    size: ElementSize = "medium"


@dataclass
class Avatar(Element):
    type: ElementType = "avatar"

    async def send(self):
        element = None

        if not self.content and not self.url and self.path:
            await self.load()

        if not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        element = self.to_dict()

        if self.emitter.emit and element:
            trace_event(f"send {self.__class__.__name__}")
            element = await self.before_emit(element)
            await self.emitter.emit("element", element)


@dataclass
class Text(Element):
    """Useful to send a text (not a message) to the UI."""

    content: Union[str, bytes] = None
    type: ElementType = "text"
    language: str = None

    async def preprocess_content(self):
        if isinstance(self.content, str):
            self.content = self.content.encode("utf-8")

    async def before_emit(self, text_element):
        if "content" in text_element and isinstance(text_element["content"], bytes):
            text_element["content"] = text_element["content"].decode("utf-8")
        return text_element


@dataclass
class Pdf(Element):
    """Useful to send a pdf to the UI."""

    type: ElementType = "pdf"


@dataclass
class Pyplot(Element):
    """Useful to send a pyplot to the UI."""

    # We reuse the frontend image element to display the chart
    type: ElementType = "image"
    size: ElementSize = "medium"

    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None

    def __post_init__(self) -> None:
        from matplotlib.figure import Figure

        if not isinstance(self.figure, Figure):
            raise TypeError("figure must be a matplotlib.figure.Figure")

        options = {
            "dpi": 200,
            "bbox_inches": "tight",
            "backend": "Agg",
            "format": "png",
        }
        image = BytesIO()
        self.figure.savefig(image, **options)
        self.content = image.getvalue()

        super().__post_init__()


class TaskStatus(Enum):
    READY = "ready"
    RUNNING = "running"
    FAILED = "failed"
    DONE = "done"


@dataclass
class Task:
    title: str = None
    status: TaskStatus = TaskStatus.READY
    forId: str = None

    def __init__(
        self,
        title: str,
        status: TaskStatus = TaskStatus.READY,
        forId: str = None,
    ):
        self.title = title
        self.status = status
        self.forId = forId

    def __post_init__(self) -> None:
        super().__post_init__()


@dataclass
class TaskList(Element):
    type = "tasklist"
    tasks: List[Task] = Field(default_factory=list, exclude=True)
    status: str = "Ready"

    def __init__(self):
        self.tasks = []
        self.content = "dummy content to pass validation"
        self.name = "tasklist"
        self.type = "tasklist"
        super().__post_init__()

    async def add_task(self, task: Task):
        self.tasks.append(task)

    async def update(self):
        await self.send()

    async def preprocess_content(self):
        # serialize enum
        tasks = [
            {"title": task.title, "status": task.status.value, "forId": task.forId}
            for task in self.tasks
        ]

        # store stringified json in content so that it's correctly stored in the database
        self.content = json.dumps(
            {
                "status": self.status,
                "tasks": tasks,
            }
        )


@dataclass
class Audio(Element):
    type: ElementType = "audio"



try:
    import langchain

    if langchain.__version__ < "0.0.198":
        raise ValueError(
            "LangChain version is too old, expected >= 0.0.198. Run `pip install langchain --upgrade`"
        )

    LANGCHAIN_INSTALLED = True
except ImportError:
    LANGCHAIN_INSTALLED = False

from chainlit.telemetry import trace
from typing import Callable, Any

from chainlit.config import config
from chainlit.utils import wrap_user_function


@trace
def langchain_factory(use_async: bool) -> Callable:
    """
    Plug and play decorator for the LangChain library.
    The decorated function should instantiate a new LangChain instance (Chain, Agent...).
    One instance per user session is created and cached.
    The per user instance is called every time a new message is received.

    Args:
        use_async bool: Whether to call the the agent asynchronously or not.

    Returns:
        Callable[[], Any]: The decorated factory function.
    """

    # Check if the factory is called with the correct parameter
    if type(use_async) != bool:
        error_message = "langchain_factory use_async parameter is required"
        raise ValueError(error_message)

    def decorator(func: Callable) -> Callable:
        config.code.lc_factory = wrap_user_function(func, with_task=True)
        return func

    config.code.lc_agent_is_async = use_async

    return decorator


@trace
def langchain_postprocess(func: Callable[[Any], str]) -> Callable:
    """
    Useful to post process the response a LangChain object instantiated with @langchain_factory.
    The decorated function takes the raw output of the LangChain object as input.
    The response will NOT be automatically sent to the UI, you need to send a Message.

    Args:
        func (Callable[[Any], str]): The post-processing function to apply after generating a response. Takes the response as parameter.

    Returns:
        Callable[[Any], str]: The decorated post-processing function.
    """

    config.code.lc_postprocess = wrap_user_function(func)
    return func


@trace
def langchain_run(func: Callable[[Any, str], str]) -> Callable:
    """
    Useful to override the default behavior of the LangChain object instantiated with @langchain_factory.
    Use when your agent run method has custom parameters.
    Takes the LangChain agent and the user input as parameters.
    The response will NOT be automatically sent to the UI, you need to send a Message.
    Args:
        func (Callable[[Any, str], str]): The function to be called when a new message is received. Takes the agent and user input as parameters and returns the output string.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """
    config.code.lc_run = wrap_user_function(func)
    return func


@trace
def langchain_rename(func: Callable[[str], str]) -> Callable[[str], str]:
    """
    Useful to rename the LangChain tools/chains used in the agent and display more friendly author names in the UI.
    Args:
        func (Callable[[str], str]): The function to be called to rename a tool/chain. Takes the original tool/chain name as parameter.

    Returns:
        Callable[[Any, str], Any]: The decorated function.
    """

    config.code.lc_rename = wrap_user_function(func)
    return func



try:
    import llama_index

    if llama_index.__version__ < "0.6.27":
        raise ValueError(
            "LlamaIndex version is too old, expected >= 0.6.27. Run `pip install llama_index --upgrade`"
        )

    LLAMA_INDEX_INSTALLED = True
except ImportError:
    LLAMA_INDEX_INSTALLED = False


from chainlit.telemetry import trace
from typing import Callable

from chainlit.config import config
from chainlit.utils import wrap_user_function


@trace
def llama_index_factory(func: Callable) -> Callable:
    """
    Plug and play decorator for the Llama Index library.
    The decorated function should instantiate a new Llama instance.
    One instance per user session is created and cached.
    The per user instance is called every time a new message is received.

    Returns:
        Callable[[], Any]: The decorated factory function.
    """

    config.code.llama_index_factory = wrap_user_function(func, with_task=True)
    return func



try:
    import langflow

    if langflow.__version__ < "0.1.4":
        raise ValueError(
            "LlamaIndex version is too old, expected >= 0.1.4. Run `pip install langflow --upgrade`"
        )

    LANGFLOW_INSTALLED = True
except ImportError:
    LANGFLOW_INSTALLED = False

from typing import Callable, Union, Dict, Optional
import aiohttp

from chainlit.telemetry import trace
from chainlit.config import config
from chainlit.lc import langchain_factory


@trace
def langflow_factory(
    use_async: bool, schema: Union[Dict, str], tweaks: Optional[Dict] = None
) -> Callable:
    """
    Plug and play decorator for the Langflow library.
    One instance per user session is created and cached.
    The per user instance is called every time a new message is received.

    Args:
        use_async bool: Whether to call the the agent asynchronously or not.
        schema (Union[Dict, str]): The langflow schema dict or url.
        tweaks Optional[Dict]: Optional tweaks to be processed


    Returns:
        Callable[[], Any]: The decorated factory function.
    """

    # Check if the factory is called with the correct parameter
    if type(schema) not in [dict, str]:
        error_message = "langflow_factory schema parameter is required"
        raise ValueError(error_message)

    # Check if the factory is called with the correct parameter
    if type(use_async) != bool:
        error_message = "langflow_factory use_async parameter is required"
        raise ValueError(error_message)

    config.code.langflow_schema = schema

    def decorator(func: Callable) -> Callable:
        async def wrapper():
            from langflow import load_flow_from_json

            schema = config.code.langflow_schema

            if type(schema) == str:
                async with aiohttp.ClientSession() as session:
                    async with session.get(
                        schema,
                    ) as r:
                        if not r.ok:
                            reason = await r.text()
                            raise ValueError(f"Error: {reason}")
                        schema = await r.json()

            flow = load_flow_from_json(input=schema, tweaks=tweaks)
            return func(flow)

        langchain_factory(use_async=use_async)(wrapper)

        return func

    return decorator



from typing import (
    Dict,
    Any,
    List,
    TypedDict,
    Optional,
    Union,
    Literal,
    TypeVar,
    Generic,
)

from abc import ABC, abstractmethod
from pydantic.dataclasses import dataclass
from dataclasses_json import dataclass_json

from chainlit.types import (
    Pagination,
    ConversationFilter,
    ElementType,
    ElementSize,
    ElementDisplay,
)


class MessageDict(TypedDict):
    conversationId: Optional[str]
    id: Optional[int]
    tempId: Optional[str]
    createdAt: Optional[int]
    content: str
    author: str
    prompt: Optional[str]
    llmSettings: Dict
    language: Optional[str]
    indent: Optional[int]
    authorIsUser: Optional[bool]
    waitForAnswer: Optional[bool]
    isError: Optional[bool]
    humanFeedback: Optional[int]


class UserDict(TypedDict):
    id: Optional[int]
    name: Optional[str]
    email: Optional[str]
    role: str


class ElementDict(TypedDict):
    id: Optional[int]
    conversationId: Optional[int]
    type: ElementType
    url: str
    name: str
    display: ElementDisplay
    size: ElementSize
    language: str
    forIds: Optional[List[Union[str, int]]]


class ConversationDict(TypedDict):
    id: Optional[int]
    createdAt: Optional[int]
    elementCount: Optional[int]
    messageCount: Optional[int]
    author: Optional[UserDict]
    messages: List[MessageDict]
    elements: Optional[List[ElementDict]]


@dataclass
class PageInfo:
    hasNextPage: bool
    endCursor: Any


T = TypeVar("T")


@dataclass_json
@dataclass
class PaginatedResponse(Generic[T]):
    pageInfo: PageInfo
    data: List[T]


class BaseAuthClient(ABC):
    user_infos: Optional[UserDict] = None
    access_token: Optional[str] = None

    @abstractmethod
    async def is_project_member(self) -> bool:
        pass

    @abstractmethod
    async def get_user_infos(self) -> UserDict:
        pass


class BaseDBClient(ABC):
    user_infos: Optional[UserDict] = None

    @abstractmethod
    async def create_user(self, variables: UserDict) -> bool:
        pass

    @abstractmethod
    async def get_project_members(self) -> List[UserDict]:
        pass

    @abstractmethod
    async def create_conversation(self) -> int:
        pass

    @abstractmethod
    async def delete_conversation(self, conversation_id: int) -> bool:
        pass

    @abstractmethod
    async def get_conversation(self, conversation_id: int) -> ConversationDict:
        pass

    @abstractmethod
    async def get_conversations(
        self, pagination: "Pagination", filter: "ConversationFilter"
    ) -> PaginatedResponse[ConversationDict]:
        pass

    @abstractmethod
    async def get_message(self, conversation_id: str, message_id: str) -> Dict:
        pass

    @abstractmethod
    async def create_message(self, variables: MessageDict) -> int:
        pass

    @abstractmethod
    async def update_message(self, message_id: int, variables: MessageDict) -> bool:
        pass

    @abstractmethod
    async def delete_message(self, message_id: int) -> bool:
        pass

    @abstractmethod
    async def upload_element(self, content: bytes, mime: str) -> str:
        pass

    @abstractmethod
    async def upsert_element(self, variables: ElementDict) -> ElementDict:
        pass

    @abstractmethod
    async def get_element(self, conversation_id: int, element_id: int) -> ElementDict:
        pass

    @abstractmethod
    async def set_human_feedback(
        self, message_id: int, feedback: Literal[-1, 0, 1]
    ) -> bool:
        pass



from typing import Union
from llama_index.response.schema import Response, StreamingResponse
from llama_index.chat_engine.types import BaseChatEngine
from llama_index.indices.query.base import BaseQueryEngine

from chainlit.message import Message
from chainlit.sync import make_async


async def run_llama(instance: Union[BaseChatEngine, BaseQueryEngine], input_str: str):
    # Trick to display the loader in the UI until the first token is streamed
    await Message(content="").send()

    response_message = Message(content="")

    if isinstance(instance, BaseQueryEngine):
        response = await make_async(instance.query)(input_str)
    elif isinstance(instance, BaseChatEngine):
        response = await make_async(instance.chat)(input_str)
    else:
        raise NotImplementedError

    if isinstance(response, Response):
        response_message.content = str(response)
        await response_message.send()
    elif isinstance(response, StreamingResponse):
        gen = response.response_gen
        for token in gen:
            await response_message.stream_token(token=token)

        if response.response_txt:
            response_message.content = response.response_txt

        await response_message.send()



import os
from chainlit.logger import logger
from chainlit.config import config, PACKAGE_ROOT

SCHEMA_PATH = os.path.join(PACKAGE_ROOT, "db/prisma/schema.prisma")


def db_push():
    from prisma.cli.prisma import run
    import prisma
    from importlib import reload

    args = ["db", "push", f"--schema={SCHEMA_PATH}"]
    env = {"LOCAL_DB_PATH": os.environ.get("LOCAL_DB_PATH")}
    run(args, env=env)

    # Without this the client will fail to initialize the first time.
    reload(prisma)


def init_local_db():
    use_local_db = config.project.database == "local"
    if use_local_db:
        if not os.path.exists(config.project.local_db_path):
            db_push()


def migrate_local_db():
    use_local_db = config.project.database == "local"
    if use_local_db:
        if os.path.exists(config.project.local_db_path):
            db_push()
            logger.info(f"Local db migrated")
        else:
            logger.info(f"Local db does not exist, skipping migration")
    else:
        logger.info(f"Database setting must be set to 'local' to migrate local db")
